import { Pulse } from '../pulse';
import { State, Group, PrimaryKey, GroupName, GroupAddOptions, Selector, Data, SelectorName } from '../internal';
export declare type DefaultDataItem = Record<string, any>;
interface RemoveOptions {
    fromGroups: (groups: string | number | Array<string>) => any;
    everywhere: () => any;
}
export declare type GroupObj<DataType> = Record<GroupName, Group<DataType>>;
export declare type SelectorObj<DataType> = Record<SelectorName, Selector<DataType>>;
export interface CollectionConfig {
    name?: string;
    primaryKey?: string | number;
    defaultGroup?: boolean;
}
export declare type Config<DataType = DefaultDataItem> = CollectionConfig | ((collection: Collection<DataType>) => CollectionConfig);
export declare class Collection<DataType extends DefaultDataItem = DefaultDataItem, G extends GroupObj<DataType> = GroupObj<DataType>, S extends SelectorObj<DataType> = SelectorObj<DataType>> {
    instance: () => Pulse;
    config: Required<CollectionConfig>;
    data: {
        [key: string]: Data<DataType>;
    };
    groups: G;
    selectors: S;
    _provisionalData: {
        [key: string]: Data<DataType>;
    };
    _provisionalGroups: {
        [key: string]: Group<DataType>;
    };
    _computedFunc?: (data: DataType) => DataType;
    _collectFunc?: (data: DataType) => DataType;
    get size(): number;
    get items(): DataType[];
    constructor(instance: () => Pulse, config: Config<DataType>);
    /**
     *  Create a group instance under this collection
     * @param groupName - The name of the group you would like to create
     * @param initialIndex - An optional array of primary keys to initialize this groups with.
     */
    createGroup<GN extends GroupName>(groupName: GN, initialIndex?: Array<PrimaryKey>): this & Collection<DataType, Record<GN, Group<DataType>>, S>;
    /**
     *  Create several group instances under this collection
     */
    createGroups<GroupNames extends GroupName>(groupNames: [GroupNames, ...GroupNames[]]): this & Collection<DataType, { [key in GroupNames]: Group<DataType>; }, S>;
    /**
     * Create a selector instance under this collection
     * @param selectorName - The name of the selector you would like to create
     * @param initialSelection - An initial PrimaryKey (string or number) to select.
     * Supports selecting data that does not yet exist, will update if that data item is eventually collected.
     */
    createSelector<SN extends SelectorName>(selectorName: SN, initialSelection?: string | number): this & Collection<DataType, G, { [key in SN]: Selector<DataType>; }>;
    /**
     *
     * @param selectorNames - An array of names to give new selectors
     * @returns The new Collection
     */
    createSelectors<SelectorNames extends SelectorName>(selectorNames: [SelectorNames, ...SelectorNames[]]): this & Collection<DataType, G, { [key in SelectorNames]: Selector<DataType>; }>;
    model(config: (...args: any) => any, options: Record<string, any>): this;
    persist(config: Record<string, string>): this;
    saveData(data: DataType, patch?: boolean): PrimaryKey | null;
    /**
     * Collect iterable data into this collection. Note:
     * - Data items must include a primary key (id)
     * @param data - Array of data, or single data object
     * @param groups - Array of group names or single group name
     */
    collect(items: DataType | DataType[], groups?: GroupName | GroupName[], config?: {
        patch?: boolean;
        method?: 'push' | 'unshift';
        forEachItem?: (item: DataType, key: PrimaryKey, index: number) => DataType;
    }): void;
    /**
     * Return an item from this collection by primaryKey as Data instance (extends State)
     * @param primaryKey - The primary key of the data
     */
    getData(id: PrimaryKey | State, options?: {
        createProvisional: boolean;
    }): Data<DataType>;
    getDataValue(id: PrimaryKey | State): DataType | null;
    /**
     * Return an group from this collection as Group instance (extends State)
     * @param groupName - The name of your group
     */
    getGroup(groupName: string): Group<DataType>;
    getGroup(groupName: keyof G): Group<DataType>;
    getGroupValue(groupName: keyof G | string): DataType[];
    /**
     * Return an group from this collection as Group instance (extends State)
     * @param groupName - The name of your group
     */
    getSelector(selectorName: keyof S): Selector<DataType>;
    getSelectorValue(selectorName: keyof S): DataType;
    private createProvisionalGroup;
    getDataValueByIndex(indexName: string, value: string): DataType;
    /**
     * Update data by id in a Pulse Collection
     * @param updateKey - The primary key of the item to update
     * @param changes - This object will be deep merged with the original
     */
    update(updateKey: PrimaryKey | State, changes?: Partial<DataType>, config?: {
        deep?: boolean;
    }): State;
    compute(func: (data: DataType) => DataType): void;
    onCollect(func: (data: DataType) => DataType): void;
    /**
     * Update data by id in a Pulse Collection
     * @param primaryKeysOrKeys - The primary key array of keys of the item(s) to update
     * @param groupNameOrNames - Group name or array of names
     */
    put(primaryKeysOrKeys: PrimaryKey | PrimaryKey[], groupNameOrNames: GroupName | GroupName[], options?: GroupAddOptions): void;
    /**
     * this is an alias function that returns other functions for removing data from a collection
     */
    remove(primaryKeysOrKeys: PrimaryKey | PrimaryKey[]): RemoveOptions;
    removeFromGroups(primaryKeyOrKeys: PrimaryKey | PrimaryKey[], groupNameOrNames: GroupName | GroupName[]): boolean;
    deleteData(primaryKeyOrKeys: PrimaryKey | PrimaryKey[], groupNameOrNames: GroupName | GroupName[]): boolean;
    private updateDataKey;
    has(primaryKey: PrimaryKey): boolean;
    rebuildGroupsThatInclude(primaryKey: PrimaryKey): void;
    getGroupsWith(primaryKey: PrimaryKey, config?: {
        includeDefault?: boolean;
    }): Group[];
    getGroupNamesWith(primaryKey: PrimaryKey): string[];
    reset(): void;
    /**
     * @deprecated Please use Collection.getData
     */
    findById(id: PrimaryKey | State): Data<DataType>;
    /**
     * @deprecated Please use Collection.getDataValue
     */
    getValueById(id: PrimaryKey | State): DataType | null;
}
export default Collection;
//# sourceMappingURL=collection.d.ts.map