"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Collection = void 0;
const internal_1 = require("../internal");
const utils_1 = require("../utils");
const deepmerge_1 = require("../helpers/deepmerge");
const utils_2 = require("../utils");
// The collection class, should be created by the Pulse class for functioning types
class Collection {
    // collection config can either be an object of type CollectionConfig or a function that returns CollectionConfig
    constructor(instance, config) {
        this.instance = instance;
        // collection data is stored here
        this.data = {};
        this.groups = {};
        this.selectors = {};
        this._provisionalData = {};
        this._provisionalGroups = {};
        if (this.instance()._collections.has(this))
            this.instance()._collections.delete(this);
        this.instance()._collections.forEach(c => {
            var _a;
            if (((_a = c === null || c === void 0 ? void 0 : c.config) === null || _a === void 0 ? void 0 : _a.name) && (config === null || config === void 0 ? void 0 : config.name)) {
                if (c.config.name === config.name)
                    this.instance()._collections.delete(c);
            }
        });
        this.instance()._collections.add(this);
        // if collection config is a function, execute and assign to config
        if (typeof config === 'function')
            config = config(this);
        // assign defaults to config object ensuring type safety
        this.config = utils_1.defineConfig(config, {
            primaryKey: 'id'
        });
        if (this.config.defaultGroup)
            this.createGroup('default');
    }
    // the amount of data items stored inside this collection
    get size() {
        return Object.keys(this.data).length;
    }
    // a getter to return the default group value
    get items() {
        var _a;
        const defaultGroup = (_a = this.groups) === null || _a === void 0 ? void 0 : _a.default;
        if (!defaultGroup)
            return [];
        return defaultGroup.output;
    }
    /**
     *  Create a group instance under this collection
     * @param groupName - The name of the group you would like to create
     * @param initialIndex - An optional array of primary keys to initialize this groups with.
     */
    createGroup(groupName, initialIndex) {
        // if (this.groups[groupName]) return this;
        const group = new internal_1.Group(() => this, initialIndex, { name: groupName });
        this.groups[groupName] = group;
        //@ts-ignore - doesn't error in vscode, but errors at build
        return this;
    }
    /**
     *  Create several group instances under this collection
     */
    createGroups(groupNames) {
        for (const name of groupNames)
            this.createGroup(name);
        //@ts-ignore - doesn't error in vscode, but errors at build
        return this;
    }
    /**
     * Create a selector instance under this collection
     * @param selectorName - The name of the selector you would like to create
     * @param initialSelection - An initial PrimaryKey (string or number) to select.
     * Supports selecting data that does not yet exist, will update if that data item is eventually collected.
     */
    createSelector(selectorName, initialSelection) {
        const selector = new internal_1.Selector(() => this, initialSelection);
        this.selectors[selectorName] = selector;
        //@ts-ignore - doesn't error in vscode, but errors at build
        return this;
    }
    /**
     *
     * @param selectorNames - An array of names to give new selectors
     * @returns The new Collection
     */
    createSelectors(selectorNames) {
        for (const name of selectorNames)
            this.createSelector(name);
        //@ts-ignore - doesn't error in vscode, but errors at build
        return this;
    }
    model(config, options) {
        return this;
    }
    persist(config) {
        return this;
    }
    // save data directly into collection storage
    saveData(data, patch) {
        let key = this.config.primaryKey;
        if (!data || !data.hasOwnProperty(key))
            return null;
        if (this._collectFunc)
            data = this._collectFunc(data);
        const existingData = this.data[data[key]];
        // if the data already exists and config is to patch, patch data
        if (patch && existingData)
            existingData.patch(data, { deep: false });
        // if already exists and no config, overwrite data
        else if (existingData)
            existingData.set(data);
        // if provisional data exists for this key, migrate data instance
        else if (this._provisionalData.hasOwnProperty(data[key])) {
            this.data[data[key]] = this._provisionalData[data[key]];
            // update provisional data instance with new data
            if (patch) {
                this.data[data[key]].patch(data, { deep: false });
            }
            else {
                this.data[data[key]].set(data);
            }
            // cleanup provisional data
            delete this._provisionalData[data[key]];
        }
        // otherwise create new data instance
        else
            this.data[data[key]] = new internal_1.Data(() => this, data);
        return data[key];
    }
    /**
     * Collect iterable data into this collection. Note:
     * - Data items must include a primary key (id)
     * @param data - Array of data, or single data object
     * @param groups - Array of group names or single group name
     */
    collect(items, groups, config = {}) {
        const _items = utils_2.normalizeArray(items);
        groups = utils_2.normalizeArray(groups);
        // is default group enabled? if so add default if not already present
        if (this.config.defaultGroup && groups.indexOf('default') === -1)
            groups.push('default');
        // if any of the groups don't already exist, create them
        for (let groupName of groups)
            !this.groups.hasOwnProperty(groupName) && this.createGroup(groupName);
        // if method is unshift reverse array order to maintain correct order
        if (config.method === 'unshift')
            _items.reverse();
        for (let [index, item] of _items.entries()) {
            if (config.forEachItem)
                item = config.forEachItem(item, item[this.config.primaryKey], index);
            let key = this.saveData(item, config.patch);
            if (key === null)
                return;
            groups.forEach(groupName => {
                let group = this.groups[groupName];
                if (!group.nextState.includes(key))
                    group.nextState[config.method || 'push'](key);
            });
        }
        for (let groupName of groups)
            this.instance().runtime.ingest(this.groups[groupName], this.groups[groupName].nextState);
    }
    /**
     * Return an item from this collection by primaryKey as Data instance (extends State)
     * @param primaryKey - The primary key of the data
     */
    getData(id, options = { createProvisional: true }) {
        if (id instanceof internal_1.State)
            id = id.value;
        if (!this.data.hasOwnProperty(id) && options.createProvisional) {
            if (this._provisionalData[id])
                return this._provisionalData[id];
            const data = new internal_1.Data(() => this, { id });
            this._provisionalData[id] = data;
            return data;
        }
        return this.data[id];
    }
    getDataValue(id) {
        var _a;
        let data = (_a = this.getData(id, { createProvisional: false })) === null || _a === void 0 ? void 0 : _a.value;
        if (!data)
            return null;
        return this._computedFunc ? this._computedFunc(data) : data;
    }
    getGroup(groupName) {
        // @TEMP fixes but with provisional (because provisional sucks)
        if (!this.groups[groupName])
            this.groups[groupName] = new internal_1.Group(() => this, [], { provisional: true, name: groupName });
        return this.groups[groupName];
        // // if provisional group exists return that
        // else if (this._provisionalGroups[groupName as GroupName]) return this._provisionalGroups[groupName as GroupName];
        // // if no group found create a provisional group
        // else return this.createProvisionalGroup(groupName as GroupName);
    }
    getGroupValue(groupName) {
        return this.getGroup(groupName).output;
    }
    /**
     * Return an group from this collection as Group instance (extends State)
     * @param groupName - The name of your group
     */
    getSelector(selectorName) {
        return this.selectors[selectorName];
    }
    getSelectorValue(selectorName) {
        return this.getSelector(selectorName).value;
    }
    createProvisionalGroup(groupName) {
        const group = new internal_1.Group(() => this, [], { provisional: true, name: groupName });
        this._provisionalGroups[groupName] = group;
        return group;
    }
    getDataValueByIndex(indexName, value) {
        return;
    }
    /**
     * Update data by id in a Pulse Collection
     * @param updateKey - The primary key of the item to update
     * @param changes - This object will be deep merged with the original
     */
    update(updateKey, changes = {}, config = {}) {
        // if State instance passed as updateKey grab the value
        if (updateKey instanceof internal_1.State)
            updateKey = updateKey.value;
        updateKey = updateKey;
        // if the primary key is changed, this will be true
        let updateDataKey = false, 
        // define aliases
        data = this.data[updateKey], primary = this.config.primaryKey;
        // if the data does not exist
        if (!this.data.hasOwnProperty(updateKey))
            return;
        // create a copy of the value for mutation
        const currentData = data.copy();
        // if the new object contains a primary key, it means we need to change the primary key
        // on the collection too, however we should defer this until after the new data is ingested into the runtime queue
        if (changes[primary])
            updateDataKey = true;
        // deep merge the new data with the existing data
        const final = config.deep ? deepmerge_1.deepmerge(currentData, changes) : utils_1.shallowmerge(currentData, changes);
        // assign the merged data to the next state of the State and ingest
        data.nextState = final;
        this.instance().runtime.ingest(data);
        // if the data key has changed move it internally and amend groups
        if (updateDataKey)
            this.updateDataKey(currentData[primary], final[primary]);
        this.rebuildGroupsThatInclude(final[primary]);
        // return the Data instance
        return this.data[final[primary]];
    }
    compute(func) {
        this._computedFunc = func;
    }
    onCollect(func) {
        this._collectFunc = func;
    }
    /**
     * Update data by id in a Pulse Collection
     * @param primaryKeysOrKeys - The primary key array of keys of the item(s) to update
     * @param groupNameOrNames - Group name or array of names
     */
    put(primaryKeysOrKeys, groupNameOrNames, options) {
        utils_2.normalizeArray(groupNameOrNames).forEach(groupName => {
            if (!this.groups.hasOwnProperty(groupName))
                this.createGroup(groupName);
            this.groups[groupName].add(primaryKeysOrKeys, options);
        });
    }
    /**
     * this is an alias function that returns other functions for removing data from a collection
     */
    remove(primaryKeysOrKeys) {
        const primaryKeys = utils_2.normalizeArray(primaryKeysOrKeys);
        return {
            fromGroups: (groups) => this.removeFromGroups(primaryKeys, groups),
            everywhere: () => this.deleteData(primaryKeys, Object.keys(this.groups))
        };
    }
    removeFromGroups(primaryKeyOrKeys, groupNameOrNames) {
        const primaryKeys = utils_2.normalizeArray(primaryKeyOrKeys);
        const groupNames = utils_2.normalizeArray(groupNameOrNames);
        groupNames.forEach(groupName => {
            if (!this.groups[groupName])
                return;
            let group = this.getGroup(groupName);
            // this loop is bad, the group should be able to handle a remove action with many keys
            primaryKeys.forEach(primaryKey => {
                group.remove(primaryKey);
            });
        });
        return true;
    }
    deleteData(primaryKeyOrKeys, groupNameOrNames) {
        const primaryKeys = utils_2.normalizeArray(primaryKeyOrKeys);
        const groupNames = utils_2.normalizeArray(groupNameOrNames);
        primaryKeys.forEach(key => {
            delete this.data[key];
            groupNames.forEach(groupName => this.groups[groupName].remove(key));
        });
        return true;
    }
    updateDataKey(oldKey, newKey) {
        // create copy of data
        const dataCopy = this.data[oldKey];
        // delete old reference
        delete this.data[oldKey];
        // apply the data in storage
        this.data[newKey] = dataCopy;
        // update groups
        for (let groupName in this.groups) {
            const group = this.getGroup(groupName);
            // if group does not contain oldKey, continue.
            if (!group._value.includes(oldKey))
                continue;
            // replace the primaryKey at current index
            group.nextState.splice(group.nextState.indexOf(oldKey), 1, newKey);
            // ingest the group
            this.instance().runtime.ingest(group);
        }
    }
    has(primaryKey) {
        return !!this.data.hasOwnProperty(primaryKey);
    }
    rebuildGroupsThatInclude(primaryKey) {
        Object.values(this.groups).forEach(group => group.rebuildOne(primaryKey));
        if (Object.keys(this._provisionalGroups).length > 0)
            Object.values(this._provisionalGroups).forEach(group => group.rebuildOne(primaryKey));
    }
    getGroupsWith(primaryKey, config = {}) {
        config = utils_1.defineConfig(config, {
            includeDefault: true
        });
        const groups = [];
        for (let key in this.groups) {
            const group = this.getGroup(key);
            if (group.has(primaryKey)) {
                if (!config.includeDefault && group.name === 'default') {
                    continue;
                }
                groups.push(group);
            }
        }
        return groups;
    }
    getGroupNamesWith(primaryKey) {
        return this.getGroupsWith(primaryKey).map(group => group.name);
    }
    reset() {
        // reset data
        this.data = {};
        // reset groups
        const groups = Object.keys(this.groups);
        groups.forEach(groupName => this.groups[groupName].reset());
        //reset selectors
        const selectors = Object.keys(this.selectors);
        selectors.forEach(selectorName => this.selectors[selectorName].reset());
    }
    /**
     * @deprecated Please use Collection.getData
     */
    findById(id) {
        return this.getData(id);
    }
    /**
     * @deprecated Please use Collection.getDataValue
     */
    getValueById(id) {
        return this.getDataValue(id);
    }
}
exports.Collection = Collection;
exports.default = Collection;
//# sourceMappingURL=collection.js.map