"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Group = void 0;
const pulse_1 = require("../pulse");
const internal_1 = require("../internal");
const utils_1 = require("../utils");
var TrackedChangeMethod;
(function (TrackedChangeMethod) {
    TrackedChangeMethod[TrackedChangeMethod["ADD"] = 0] = "ADD";
    TrackedChangeMethod[TrackedChangeMethod["REMOVE"] = 1] = "REMOVE";
    TrackedChangeMethod[TrackedChangeMethod["UPDATE"] = 2] = "UPDATE";
})(TrackedChangeMethod || (TrackedChangeMethod = {}));
class Group extends internal_1.State {
    constructor(context, initialIndex, config = {}) {
        // This invokes the parent class with either the collection or the Pulse instance as context
        // This means groups can be created before (or during) a Collection instantiation
        super((context() instanceof pulse_1.Pulse ? context : context().instance), initialIndex || []);
        // group state
        this._output = [];
        this._preciseIndex = [];
        this._missingPrimaryKeys = [];
        if (context() instanceof internal_1.Collection)
            this.collection = context;
        if (config.name)
            this.key(config.name);
        if (config.provisional)
            this.provisional = config.provisional;
        this.type(Array);
        this.sideEffects = () => {
            if (config.lazyBuild != false)
                this._outdated = true;
            else
                this.build();
        };
        // initial build
        if (config.lazyBuild != false)
            this._output = [];
        else
            this.build();
    }
    // getter for precise index
    get index() {
        return this._preciseIndex;
    }
    get size() {
        return this._preciseIndex.length;
    }
    // getter for group output, contains built collection data
    get output() {
        if (this._outdated)
            this.build();
        if (this.instance().runtime.trackState)
            this.instance().runtime.foundState.add(this);
        return this._output;
    }
    build() {
        if (!Array.isArray(this._value))
            return [];
        // If this group has been modified at specific indexes, avoid rebuilding entire group data
        if (this._trackedIndexChanges) {
            for (const change of this._trackedIndexChanges) {
                // if tracked change is to remove the item at this index
                if (change.method == TrackedChangeMethod.REMOVE) {
                    this._preciseIndex.splice(change.index, 1);
                    this._output.splice(change.index, 1);
                    continue;
                }
                //
                else {
                    const data = this.getCollectionData(change.key, change.index);
                    if (data && change.method == TrackedChangeMethod.ADD) {
                        this._output.splice(change.index, 0, this.renderData(data));
                        this._preciseIndex.splice(change.index, 0, change.key);
                    }
                    //
                    else if (data && change.method == TrackedChangeMethod.UPDATE) {
                        this._output[change.index] = this.renderData(data);
                    }
                }
            }
            // clean up
            delete this._trackedIndexChanges;
        }
        // Build group output from scratch
        else {
            this._missingPrimaryKeys = [];
            this._preciseIndex = [];
            // build the group data from collection data using the index
            this._output = this._value
                .map((primaryKey, index) => {
                const data = this.getCollectionData(primaryKey, index);
                if (data) {
                    this._preciseIndex.push(primaryKey);
                    return this.renderData(data);
                }
                return undefined;
            })
                .filter(item => item != undefined);
        }
        delete this._outdated;
    }
    getCollectionData(key, index) {
        let data = this.collection().data[key];
        if (!data) {
            this._missingPrimaryKeys.push(key);
            return undefined;
        }
        return data;
    }
    renderData(data) {
        if (this.computedFunc) {
            let dataComputed = this.computedFunc(data.copy());
            return dataComputed;
            // use collection level computed func if local does not exist
        }
        else if (this.collection()._computedFunc) {
            let dataComputed = this.collection()._computedFunc(data.copy());
            return dataComputed;
        }
        return data.getPublicValue();
    }
    has(primaryKey) {
        return this.value.includes(primaryKey) || false;
    }
    compute(func) {
        this.computedFunc = func;
    }
    add(primaryKeyOrKeys, options = {}) {
        options = utils_1.defineConfig(options, { method: 'push', overwrite: true, softRebuild: true });
        // set defaults
        let value = this.copy();
        const useIndex = options.atIndex != undefined;
        for (let [i, primaryKey] of utils_1.normalizeArray(primaryKeyOrKeys).entries()) {
            const exists = value.includes(primaryKey);
            if (options.overwrite)
                value = value.filter(i => i !== primaryKey);
            // if we do not want to overwrite and key already exists in group, exit
            else if (exists)
                return this;
            // if atIndex is set, inject at that index.
            if (useIndex) {
                // if index is greater than length insert at the end of the array
                if (options.atIndex > value.length)
                    options.atIndex = value.length - 1;
                value.splice(options.atIndex + i, 0, primaryKey);
            }
            // push or unshift into state
            else
                value[options.method](primaryKey);
            if (options.softRebuild) {
                this.trackChange({
                    method: exists ? TrackedChangeMethod.UPDATE : TrackedChangeMethod.ADD,
                    key: primaryKey,
                    index: useIndex ? options.atIndex : options.method == 'push' ? value.length - 1 : 0
                });
            }
        }
        this.set(value, { _caller: this.add });
        return this;
    }
    remove(primaryKey, options = {}) {
        options = utils_1.defineConfig(options, { softRebuild: true });
        const value = this.copy();
        const index = this._preciseIndex.indexOf(primaryKey);
        if (index == -1)
            return this;
        if (options.softRebuild)
            this.trackChange({ index, key: primaryKey, method: TrackedChangeMethod.REMOVE });
        value.splice(index, 1);
        this.set(value, { _caller: this.remove });
        return this;
    }
    trackChange(change) {
        if (!this._trackedIndexChanges)
            this._trackedIndexChanges = [];
        this._trackedIndexChanges.push(change);
    }
    rebuildOne(primaryKey) {
        const index = this._preciseIndex.indexOf(primaryKey);
        if (index == -1)
            return this;
        this.trackChange({ index, key: primaryKey, method: TrackedChangeMethod.UPDATE });
        this.set();
    }
}
exports.Group = Group;
exports.default = Group;
//# sourceMappingURL=group.js.map