"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Event = void 0;
// Event class
class Event {
    constructor(instance, config = {}) {
        this.instance = instance;
        this.config = config;
        // store the callbacks as a set of functions
        this.callbacks = new Set();
        // initiate uses state if applicable
        if (config.disableAfterUses)
            this.uses = 0;
        if (config.queue)
            this.queue = [];
    }
    // register subscribers
    on(callback) {
        // on returns a clean up function, defined here for multiple return points
        const cleanupFunc = () => this.unsub(callback);
        // if maxSubs is set, and there are less than or equal to the maximum, fail silently.
        if (this.config.maxSubs !== undefined && this.callbacks.size <= this.config.maxSubs)
            return cleanupFunc;
        // if destroy after uses is truthy and the uses is less than or equal the destroy amount
        if (this.config.disableAfterUses && this.uses > this.config.disableAfterUses) {
            this.disable();
            return cleanupFunc;
        }
        // add the callback to Event callback set and return cleanup function
        this.callbacks.add(callback);
        return cleanupFunc;
    }
    // run all the callbacks in this event and pass the payload
    emit(payload) {
        // if Event is disabled block emitting
        if (this.config.enabled !== undefined && !this.config.enabled)
            return;
        if (this.config.throttle) {
            this.handleThrottle(payload);
        }
        else {
            this.emitter(payload);
        }
    }
    disable() {
        this.config.enabled = false;
    }
    onNext(callback) {
        this.onNextCallback = callback;
    }
    // Private functions
    emitter(payload) {
        // foreach callback, invoke the saved function
        this.callbacks.forEach(callback => callback(payload));
        if (typeof this.onNextCallback === 'function') {
            this.onNextCallback(payload);
            delete this.onNextCallback;
        }
        // increment the uses if
        if (this.uses !== undefined)
            this.uses++;
    }
    unsub(callback) {
        this.callbacks.delete(callback);
    }
    handleThrottle(payload) {
        const throttling = this.currentTimeout !== undefined;
        // throttling with a queue? push to queue and reset timeout
        if (throttling && this.queue) {
            this.queue.push(payload);
            clearTimeout(this.currentTimeout);
            this.currentTimeout = undefined;
        }
        // throttling without a queue? exit
        else if (throttling)
            return;
        // throttle is not running, begin timeout chain
        else {
            const looper = (payload) => {
                this.currentTimeout = setTimeout(() => {
                    this.currentTimeout = undefined;
                    // emit the event with passed in payload
                    this.emitter(payload);
                    // if using the queue grab the next payload from the queue and loop back with a new timer
                    if (this.queue && this.queue.length > 0)
                        looper(this.queue.shift());
                }, this.config.throttle);
            };
            looper(payload);
        }
        //
        return;
    }
}
exports.Event = Event;
//# sourceMappingURL=event.js.map