"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.route = exports.Route = void 0;
const ensureProperHeaders = headers => {
    let obj = {};
    Object.keys(headers).forEach(t => {
        const formattedHeader = t
            .toLowerCase()
            .split('-')
            .map(token => token.charAt(0).toUpperCase() + token.slice(1))
            .join('-');
        obj[formattedHeader] = headers[t];
    });
    return obj;
};
class Route {
    constructor(config = { options: {} }) {
        this.config = config;
        // ensure headers are in the correct format
        if (config.options && config.options.headers) {
            config.options.headers = ensureProperHeaders(config.options.headers);
        }
        // ensure the fetch options are present
        if (!config.options)
            config.options = {};
    }
    get(endpoint) {
        return this.send('GET', endpoint);
    }
    post(endpoint, payload) {
        return this.send('POST', endpoint, payload);
    }
    put(endpoint, payload) {
        return this.send('PUT', endpoint, payload);
    }
    patch(endpoint, payload) {
        return this.send('PATCH', endpoint, payload);
    }
    delete(endpoint, payload) {
        return this.send('DELETE', endpoint, payload);
    }
    send(method, endpoint, payload) {
        var _a, _b, _c, _d, _e;
        return __awaiter(this, void 0, void 0, function* () {
            // initial definitions
            let fullUrl, response, config = Route.cloneDefaultConfig(this.config);
            // inject method into request options
            config.options.method = method;
            if (!config.options.headers)
                config.options.headers = {};
            // inject payload into request body
            if (typeof payload === 'object') {
                // stringify body
                config.options.body = JSON.stringify(payload);
                // if body is object and no content type specified, auto assign content-type to application/json
                if (!config.options.headers['Content-Type'])
                    config.options.headers['Content-Type'] = 'application/json';
            }
            else {
                config.options.body = payload;
            }
            // construct endpoint
            if (endpoint.startsWith('http'))
                fullUrl = endpoint;
            else
                fullUrl = `${this.config.baseURL ? this.config.baseURL : ''}/${endpoint}`;
            // make the request
            let timedout = false;
            if (config.timeout) {
                let t;
                const timeout = new Promise(resolve => {
                    t = setTimeout(() => {
                        timedout = true;
                        resolve();
                    }, config.timeout);
                });
                const request = new Promise((resolve, reject) => {
                    fetch(fullUrl, config.options)
                        .then(data => {
                        clearTimeout(t);
                        resolve(data);
                    })
                        .catch(reject);
                });
                response = yield Promise.race([timeout, request]);
            }
            else {
                response = yield fetch(fullUrl, config.options);
            }
            // if we got here, PulseResponse is the actual response object
            let res = {
                status: timedout ? 408 : (_a = response) === null || _a === void 0 ? void 0 : _a.status,
                raw: response,
                data: {},
                type: ((_c = (_b = response) === null || _b === void 0 ? void 0 : _b.headers) === null || _c === void 0 ? void 0 : _c.get('content-type')) || 'text/plain',
                timedout
            };
            // extract response data
            if ((_d = res.type) === null || _d === void 0 ? void 0 : _d.includes('application/json')) {
                res.data = yield res.raw.json();
            }
            else if (typeof ((_e = res === null || res === void 0 ? void 0 : res.raw) === null || _e === void 0 ? void 0 : _e.text) === 'function') {
                res.data = yield res.raw.text();
            }
            return res;
        });
    }
    static cloneDefaultConfig(config) {
        const cfg = Object.assign({}, config);
        cfg.options = Object.assign({}, config.options);
        cfg.options.headers = Object.assign({}, config.options.headers);
        return cfg;
    }
}
exports.Route = Route;
/**
 * @param config.options normal fetch options such as headers, credentials, etc.
 * @param config.baseURL The URL to be used on each request (if left empty, defaults to current hostname)
 * @returns The configured route function
 */
function route(config) {
    // if no baseURL is provided, use the current hostname
    if (typeof config.baseURL !== 'string') {
        config.baseURL = '';
    }
    if (config.baseURL.endsWith('/')) {
        config.baseURL = config.baseURL.substring(0, config.baseURL.length);
    }
    const route = new Route({
        options: config.options,
        baseURL: config.baseURL,
        timeout: config.timeout || undefined // this is just incease the user passes 0, it should be treated as undefined
    });
    /**
     * @param path The path to be appended to the baseURL
     * @param config.params The params to be appended to the path
     * @param config.query The query params to be appended to the POST path
     * @param config.body The body to be sent with the request
     * @param config.method The method to be used for the request
     * @param config.options The fetch options to be used for the request
     */
    return (path, inConfig) => __awaiter(this, void 0, void 0, function* () {
        // if(inConfig.path.startsWith('/')){inConfig.path = inConfig.path.substring(1)}
        try {
            if (path.startsWith('/')) {
                path = path.substring(1);
            }
            if (inConfig.options) {
                route.config.options = Object.assign(Object.assign({}, route.config.options), inConfig.options);
            }
            if (!inConfig.method) {
                inConfig.method = 'GET';
            }
            else {
                inConfig.method = inConfig.method.toUpperCase();
            }
            const query = inConfig.query
                ? Object.keys(inConfig.query)
                    .map(key => `${key}=${encodeURIComponent(inConfig.query[key])}`)
                    .join('&')
                : null;
            const params = inConfig.params
                ? Object.keys(inConfig.params)
                    .map(key => `${key}=${encodeURIComponent(inConfig.params[key])}`)
                    .join('&')
                : null;
            switch (inConfig.method) {
                case 'PATCH':
                    return yield route.send('PATCH', path, inConfig.body);
                case 'POST':
                    return yield route.send('POST', `${path}${params ? `?${params}` : ''}`, inConfig.body);
                case 'PUT':
                    return yield route.send("PUT", path, inConfig.body);
                default:
                    return yield route.send(inConfig.method, `${path}${query ? `?${query}` : ''}`);
            }
        }
        catch (e) {
            // throw e;
            return Promise.reject(e);
        }
    });
}
exports.route = route;
//# sourceMappingURL=route.js.map