/// <reference types="node" />
import { Pulse } from './pulse';
import { Dep } from './internal';
export declare class State<ValueType = any> {
    instance: () => Pulse;
    initialState?: ValueType | null;
    name?: string;
    _value: ValueType;
    get value(): ValueType;
    dep: Dep;
    watchers?: {
        [key: string]: any;
    };
    previousState: ValueType;
    nextState: ValueType;
    isSet: boolean;
    persistState: boolean;
    typeOfVal?: string;
    enableHistory?: boolean;
    history?: HistoryItem<ValueType>[];
    sideEffects?: Function;
    intervalId?: NodeJS.Timer | number;
    set mutable(value: ValueType);
    get mutable(): ValueType;
    get exists(): boolean;
    constructor(instance: () => Pulse, initialState?: ValueType | null, deps?: Array<Dep>);
    /**
     * Directly set state to a new value, if nothing is passed in State.nextState will be used as the next value
     * @param newState - The new value for this state
     */
    set(newState?: ValueType | SetFunc<ValueType>, options?: {
        background?: boolean;
        _caller?: Function;
    }): this;
    getPublicValue(): ValueType;
    patch(targetWithChange: any, config?: {
        deep?: boolean;
    }): this;
    /**
     * On a certain interval of milliseconds, set the state's value to the return value of a provided callback.
     * @param setFunc Function that returns the next value to be applied to state.
     * @param ms Time, in milliseconds, between individual runs of the interval.
     * @returns Native handle from `setInterval` that can be passed into `clearInterval`.
     */
    interval(setFunc: (currentValue: ValueType) => any, ms?: number): this;
    clearInterval(): void;
    persist(key?: string): this;
    /**
     * @public
     * Create a watcher that will fire a callback then destroy itself after invoking
     */
    onNext(callback: (value: ValueType) => void): void;
    /**
     * @public
     * Set a name for this State, required to persist
     */
    key(key: string): this;
    /**
     * @public
     * Fix the type to one of 'String', 'Boolean', 'Array', 'Object' or 'Number'
     */
    type(type: TypeString | TypeConstructor): this;
    /**
     * @public
     * Watch state for changes, run callback on each change
     */
    watch(callback: Callback<ValueType>): string | number;
    watch(key: string | number, callback: Callback<ValueType>): this;
    /**
     * @public
     * Remove watcher by key
     */
    removeWatcher(key: string | number): this;
    /**
     * @public
     * Restore previous state
     */
    undo(): void;
    /**
     * @public
     * Records all state changes
     */
    record(record?: boolean): this;
    /**
     * @public
     * If State is boolean, invert
     */
    toggle(): this;
    /**
     * @public
     * Reset the State to as declared
     */
    reset(): this;
    /**
     * @public
     * Returns a copy of the current value, objects and arrays will be cloned
     */
    copy(): ValueType;
    /**
     * @public
     * Is the value equal to parameter
     */
    is(x: any): boolean;
    /**
     * @public
     * Is the value not equal to parameter
     */
    isNot(x: any): boolean;
    /**
     * @internal
     * Write value directly to State
     */
    privateWrite(value: ValueType): void;
    /**
     * @internal
     *
     */
    private isCorrectType;
    /**
     * @internal
     *
     */
    destroy(): void;
    /**
     * @internal
     *
     */
    getPersistableValue(): any;
}
export declare type StateGroupDefault = {
    [key: string]: State | any;
};
export declare const StateGroup: (instance: () => Pulse, stateGroup: Object) => any;
export default State;
export declare type SetFunc<ValueType> = (state: ValueType) => ValueType;
declare type TypeString = 'string' | 'boolean' | 'array' | 'object' | 'number';
declare type TypeConstructor = StringConstructor | BooleanConstructor | ArrayConstructor | ObjectConstructor | NumberConstructor;
export interface HistoryItem<ValueType = any> {
    previousValue: ValueType;
    value: ValueType;
    timestamp: Date;
    name?: string;
}
declare type Callback<T = any> = (value: T) => void;
//# sourceMappingURL=state.d.ts.map