"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.StateGroup = exports.State = void 0;
const internal_1 = require("./internal");
const utils_1 = require("./utils");
const deepmerge_1 = require("./helpers/deepmerge");
class State {
    constructor(instance, initialState, deps = []) {
        this.instance = instance;
        this.initialState = initialState;
        // internal storage for the current value
        this._value = null;
        // the previous value of this State
        this.previousState = null;
        // a copy of the current value that will be used if no param is passed on State.set()
        this.nextState = null;
        // if the value has been changed from initial value
        this.isSet = false;
        if (this.instance()._state.has(this))
            this.instance()._state.delete(this);
        this.instance()._state.forEach((s) => {
            if ((s === null || s === void 0 ? void 0 : s.name) && this.name) {
                if (s.name === this.name)
                    this.instance()._state.delete(s);
            }
        });
        this.instance()._state.add(this);
        // initialize the dependency manager
        this.dep = new internal_1.Dep(deps, () => this);
        // write the initial value to this State
        this.privateWrite(initialState === undefined ? null : initialState);
    }
    // getter for the current value
    get value() {
        if (this.instance().runtime.trackState)
            this.instance().runtime.foundState.add(this);
        return this._value;
    }
    // // for extended classes to store a derived value, such as Group
    // public output?: any;
    // getter and setter for the State value, best for I/O binding
    set mutable(value) {
        this.set(value);
    }
    get mutable() {
        return this.value;
    }
    // is value truthey or falsey
    get exists() {
        return !!this.value;
    }
    /**
     * Directly set state to a new value, if nothing is passed in State.nextState will be used as the next value
     * @param newState - The new value for this state
     */
    set(newState, options = {}) {
        // if newState not provided, just ingest update with existing value
        if (newState === undefined) {
            this.instance().runtime.ingest(this, undefined, { jobSpawnedFrom: options._caller });
            return this;
        }
        // if newState is a function, run that function and supply existing value as first param
        if (typeof newState === 'function')
            newState = newState(this._value);
        // check type if set and correct otherwise exit
        if (this.typeOfVal && !this.isCorrectType(newState)) {
            console.warn(`Pulse: Error setting state: Incorrect type (${typeof newState}) was provided. Type fixed to ${this.typeOfVal}`);
            return this;
        }
        // ingest update using most basic mutation method
        if (options.background) {
            this.privateWrite(newState);
            if (this.sideEffects)
                this.sideEffects();
        }
        else {
            this.instance().runtime.ingest(this, newState, { jobSpawnedFrom: options._caller });
        }
        this.isSet = true;
        return this;
    }
    getPublicValue() {
        if (this['output'] !== undefined)
            return this['output'];
        return this._value;
    }
    patch(targetWithChange, config = {}) {
        if (!(typeof this._value === 'object'))
            return this;
        this.nextState = config.deep === false ? utils_1.shallowmerge(this.nextState, targetWithChange) : deepmerge_1.deepmerge(this.nextState, targetWithChange);
        this.set();
        return this;
    }
    /**
     * On a certain interval of milliseconds, set the state's value to the return value of a provided callback.
     * @param setFunc Function that returns the next value to be applied to state.
     * @param ms Time, in milliseconds, between individual runs of the interval.
     * @returns Native handle from `setInterval` that can be passed into `clearInterval`.
     */
    interval(setFunc, ms) {
        if (this.intervalId !== undefined)
            return this;
        this.intervalId = setInterval(() => {
            this.set(setFunc(this.value));
        }, ms !== null && ms !== void 0 ? ms : 1000);
        return this;
    }
    clearInterval() {
        if (this.intervalId) {
            clearInterval(this.intervalId);
            delete this.intervalId;
        }
    }
    persist(key) {
        this.persistState = true;
        this.instance().storage.handleStatePersist(this, key);
        return this;
    }
    /**
     * @public
     * Create a watcher that will fire a callback then destroy itself after invoking
     */
    onNext(callback) {
        if (!this.watchers)
            this.watchers = {};
        this.watchers['_on_next_'] = () => {
            callback(this.getPublicValue());
            delete this.watchers['_on_next_'];
        };
    }
    /**
     * @public
     * Set a name for this State, required to persist
     */
    key(key) {
        // if persist was attempted before, but no key was provided retry persist function
        if (!this.name && this.persistState)
            this.persist(key);
        this.name = key;
        return this;
    }
    /**
     * @public
     * Fix the type to one of 'String', 'Boolean', 'Array', 'Object' or 'Number'
     */
    type(type) {
        const supportedConstructors = ['String', 'Boolean', 'Array', 'Object', 'Number'];
        if (typeof type === 'function' && supportedConstructors.includes(type.name))
            this.typeOfVal = type.name.toLowerCase();
        else if (typeof type === 'string' && supportedConstructors.map(i => i.toLowerCase()).includes(type))
            this.typeOfVal = type.toLowerCase();
        return this;
    }
    watch(keyOrCallback, callback) {
        if (!this.watchers)
            this.watchers = {};
        let genKey = typeof keyOrCallback === 'function', key;
        if (genKey) {
            key = this.instance().getNonce();
            callback = keyOrCallback;
        }
        else
            key = keyOrCallback;
        this.watchers[key] = callback;
        return genKey ? key : this;
    }
    /**
     * @public
     * Remove watcher by key
     */
    removeWatcher(key) {
        if (this.watchers[key])
            delete this.watchers[key];
        return this;
    }
    /**
     * @public
     * Restore previous state
     */
    undo() {
        this.set(this.previousState);
    }
    /**
     * @public
     * Records all state changes
     */
    record(record) {
        if (!this.history)
            this.history = [];
        this.enableHistory = record === false ? false : true;
        return this;
    }
    /**
     * @public
     * If State is boolean, invert
     */
    toggle() {
        if (typeof this._value === 'boolean') {
            // @ts-ignore
            this.set(!this._value);
        }
        return this;
    }
    /**
     * @public
     * Reset the State to as declared
     */
    reset() {
        this.isSet = false;
        this.previousState = null;
        if (this.persistState)
            this.instance().storage.remove(this.name);
        this.instance().runtime.ingest(this, this.initialState);
        return this;
    }
    /**
     * @public
     * Returns a copy of the current value, objects and arrays will be cloned
     */
    copy() {
        return utils_1.copy(this.value);
    }
    /**
     * @public
     * Is the value equal to parameter
     */
    is(x) {
        return this.value === x;
    }
    /**
     * @public
     * Is the value not equal to parameter
     */
    isNot(x) {
        return this.value !== x;
    }
    /**
     * @internal
     * Write value directly to State
     */
    privateWrite(value) {
        if (this.enableHistory) {
            this.history.push({
                value: value,
                previousValue: this._value,
                timestamp: new Date()
            });
        }
        if (this.instance().config.globalHistory) {
            this.instance().history.push({
                value: value,
                previousValue: this._value,
                timestamp: new Date(),
                name: this.name
            });
        }
        this._value = utils_1.copy(value);
        this.nextState = utils_1.copy(value);
        // If
        if (this.persistState)
            this.instance().storage.set(this.name, this.getPersistableValue());
    }
    /**
     * @internal
     *
     */
    isCorrectType(value) {
        let type = typeof value;
        if (type === 'object' && Array.isArray(value))
            type = 'array';
        return type === this.typeOfVal;
    }
    /**
     * @internal
     *
     */
    destroy() {
        this.dep.deps.clear();
        this.dep.subs.clear();
    }
    /**
     * @internal
     *
     */
    getPersistableValue() {
        return this.value;
    }
}
exports.State = State;
exports.StateGroup = (instance, stateGroup) => {
    let group = {};
    for (let name in stateGroup) {
        group[name] = new State(instance, stateGroup[name]);
        group[name].name = name;
    }
    return group;
};
exports.default = State;
//# sourceMappingURL=state.js.map