"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.persist = exports.Storage = void 0;
const utils_1 = require("./utils");
class Storage {
    constructor(instance, config) {
        this.instance = instance;
        this.storageReady = false;
        this.persistedState = new Set();
        this.config = utils_1.defineConfig(config, {
            prefix: 'pulse',
            type: 'localStorage'
        });
        // assume if user provided get, set or remove methods that the storage type is custom
        if (this.config.get || this.config.set || this.config.remove)
            this.config.type = 'custom';
        const ls = this.getLocalStorage();
        if (this.config.type === 'localStorage' && ls) {
            // assign localStorage crud functions to config object
            ['get', 'set', 'remove'].forEach(type => (this.config[type] = ls[`${type}Item`].bind(ls)));
            if (this.config.watch && typeof window !== "undefined") {
                try {
                    window.addEventListener('storage', (event) => {
                        const k = this.extractKey(event.key);
                        const states = this.persistedState.values();
                        for (const st of states) {
                            if (st.name === k) {
                                let nv;
                                try {
                                    nv = JSON.parse(event.newValue);
                                }
                                catch (e) {
                                    nv = event.newValue;
                                }
                                st.set(nv);
                            }
                        }
                    });
                }
                catch (error) {
                    // quietly fail
                }
            }
            this.storageReady = true;
        }
        else {
            // Local storage not available, fallback to custom.
            this.config.type = 'custom';
            // ensuring all required storage properties are set
            if (isFunction(this.config.get) && isFunction(this.config.set) && isFunction(this.config.remove)) {
                // if asynchronous and developer did not explicitly define so, check
                if (this.config.async === undefined && utils_1.isAsync(this.config.get))
                    this.config.async = true;
                this.storageReady = true;
            }
            else {
                // console.warn('Pulse Error: Persistent storage not configured, check get, set and remove methods', this.config);
                this.storageReady = false;
            }
        }
    }
    get(key) {
        if (!this.storageReady)
            return;
        if (this.config.async) {
            return new Promise((resolve, reject) => {
                this.config
                    .get(this.getKey(key))
                    .then(res => {
                    // if result is not JSON for some reason, return it.
                    if (typeof res !== 'string')
                        return resolve(res);
                    resolve(JSON.parse(res));
                })
                    .catch(reject);
            });
        }
        else {
            try {
                return JSON.parse(this.config.get(this.getKey(key)));
            }
            catch (e) {
                return undefined;
            }
        }
    }
    set(key, value) {
        if (!this.storageReady)
            return;
        this.config.set(this.getKey(key), JSON.stringify(value));
    }
    remove(key) {
        if (!this.storageReady)
            return;
        this.config.remove(this.getKey(key));
    }
    getKey(key) {
        return `_${this.config.prefix}_${key}`;
    }
    extractKey(key) {
        return key.slice(this.config.prefix.length + 1);
    }
    // used by State and Selector to persist value inside storage
    handleStatePersist(state, key) {
        const storage = this;
        // validation
        if (!key && state.name) {
            key = state.name;
        }
        else if (!key) {
            return;
            // console.warn('Pulse Persist Error: No key provided');
        }
        else {
            state.name = key;
        }
        // add ref to state instance inside storage
        storage.persistedState.add(state);
        // handle the value
        const handle = (storageVal) => {
            // if no storage value found, set current value in storage
            if (storageVal === null)
                storage.set(state.name, state.getPersistableValue());
            // if Selector, select current storage value
            else if (typeof state['select'] === 'function' && (typeof storageVal === 'string' || typeof storageVal === 'number'))
                state['select'](storageVal);
            // otherwise just ingest the storage value so that the State updates
            else
                state.instance().runtime.ingest(state, storageVal);
        };
        // Check if promise, then handle value
        if (storage.config.async)
            storage.get(state.name).then((value) => handle(value));
        // non promise
        else
            handle(storage.get(state.name));
    }
    getLocalStorage() {
        try {
            const ls = (window === null || window === void 0 ? void 0 : window.localStorage) ? window.localStorage : localStorage;
            if (typeof ls.getItem !== 'function')
                return false;
            return ls;
        }
        catch (e) {
            return false;
        }
    }
}
exports.Storage = Storage;
function isFunction(func) {
    return typeof func === 'function';
}
exports.default = Storage;
// Handy utils
function persist(items) {
    items.forEach(item => item.persist(item.name));
}
exports.persist = persist;
//# sourceMappingURL=storage.js.map